#ifndef HIST_HUE_SAT_HPP
#define HIST_HUE_SAT_HPP

#include "opencv2/opencv.hpp"

class HistHueSat {
public:
    HistHueSat(int hbin = 64, int sbin = 64) : hbin(hbin), sbin(sbin) {
        hist = cv::Mat(sbin, hbin, CV_32F);
        clear();
    }
    float coeffBhattacharyya(const HistHueSat& rhs) const {
        cv::Mat sqrt_of_mult;
        cv::sqrt(hist.mul(rhs.hist), sqrt_of_mult);
        return (float)cv::sum(sqrt_of_mult)[0];
    }

    void calc(const cv::Mat& hsv_img) {
        cv::RotatedRect trect;
        trect.size = hsv_img.size();
        trect.center.x = hsv_img.cols / 2.0f;
        trect.center.y = hsv_img.rows / 2.0f;
        calc(hsv_img, trect, false);
    }
    void calc(const cv::Mat& hsv_img, const cv::RotatedRect& trect,
              bool is_weighted = true) {

        int ww = (int)(trect.size.width / 2);
        int hh = (int)(trect.size.height / 2);
        int cx = (int)(trect.center.x);
        int cy = (int)(trect.center.y);
        int xmin = cv::max(cx - ww, 0);
        int xmax = cv::min(cx + ww, hsv_img.cols - 1);
        int ymin = cv::max(cy - hh, 0);
        int ymax = cv::min(cy + hh, hsv_img.rows - 1);
        float w_normalizer = 2 / trect.size.width;
        float h_normalizer = 2 / trect.size.height;

        clear();
        for (int j = ymin; j <= ymax; j++) {
            for (int i = xmin; i <= xmax; i++) {
                float x_norm = (i - cx) * w_normalizer;
                float y_norm = (j - cy) * h_normalizer;
                float r2 = x_norm * x_norm + y_norm * y_norm;
                float weight;

                if (is_weighted) { // Epanechnikov profile (up to scale factor)
                    weight = (r2 <= 1)? 1 - r2 : 0.0f;
                } else {
                    weight = 1.0f;
                }

                int hue = hsv_img.at<cv::Vec3b>(j, i)[0];
                int sat = hsv_img.at<cv::Vec3b>(j, i)[1];
                vote(weight, hue, sat);
            }
        }
        normalize();
    }
    void visualizeLegend(cv::Mat& disp) {
        visualize(disp, true);
    }
    void visualize(cv::Mat& disp, bool is_legend = false) {
        float max = calcMax();
        int width = disp.cols;
        int height = disp.rows;
        for (int j = 0; j < height; j++) {
            for (int i = 0; i < width; i++) {
                int hue = (i * 180) / width;
                int sat = (j * 256) / height;
                int value;
                if (is_legend) {
                    value = 255;
                } else {
                    /* linear mapping */
                    //value = (int)((at(hue, sat) / max) * 255);
                    /* logarithmic mapping */
                    value = (int)(365 * log(at(hue, sat) / max + 1.0f));

                    value = cv::min(value, 255);
                }
                disp.at<cv::Vec3b>(j, i) = cv::Vec3b(hue, sat, value);
            }
        }
        cv::cvtColor(disp, disp, CV_HSV2BGR);
    }
    float at(int hue, int sat) const {
        int h_idx = (hue * hbin) / 180;
        int s_idx = (sat * sbin) / 256;
        return hist.at<float>(s_idx, h_idx);
    }

private:
    void setAt(float value, int hue, int sat) {
        int h_idx = (hue * hbin) / 180;
        int s_idx = (sat * sbin) / 256;
        hist.at<float>(s_idx, h_idx) = value;
    }
    void clear() {
        hist = 0.0f;
    }
    float calcMax() {
        float max = 0.0f;
        for (int s = 1; s < sbin - 1; s++) { // smallest and largest sat ignored
            for (int h = 0; h < hbin; h++) {
                float val = hist.at<float>(s, h);
                if (val > max) {
                    max = val;
                }
            }
        }
        return max;
    }
    float calcSum() {
        cv::Scalar scalar_sum = cv::sum(hist);
        return (float)scalar_sum[0];
    }
    void normalize() {
        hist /= calcSum();
    }
    void vote(float weight, int hue, int sat) {
        float histval = at(hue, sat);
        histval += weight;
        setAt(histval, hue, sat);
    }

    cv::Mat hist;
    int hbin;
    int sbin;
};

#endif /* HIST_HUE_SAT_HPP */
