#include "opencv2/opencv.hpp"

#include "../hist_hue_sat.hpp"
#include "../mouse_rect.hpp"

void printDeltaMilliSec(double tick_begin, double tick_end)
{
    double delta = 1000.0 * (tick_end - tick_begin) / cv::getTickFrequency();
    printf("%f [ms]\n", delta);
}

float meanShift(const cv::Mat& hsv_img, cv::RotatedRect& trect,
                const HistHueSat& hist_model,
                HistHueSat& hist_cand)
{
    double tick_begin = (double)cv::getTickCount();

    int ww = (int)(trect.size.width / 2);
    int hh = (int)(trect.size.height / 2);
    float w_normalizer = 2 / trect.size.width;
    float h_normalizer = 2 / trect.size.height;

    const int max_iter = 20;
    int iter = 0;
    float displace2;

    do {
        hist_cand.calc(hsv_img, trect);

        int cx = (int)(trect.center.x);
        int cy = (int)(trect.center.y);
        int xmin = cv::max(cx - ww, 0);
        int xmax = cv::min(cx + ww, hsv_img.cols - 1);
        int ymin = cv::max(cy - hh, 0);
        int ymax = cv::min(cy + hh, hsv_img.rows - 1);
        float m0 = 0.0f, mx = 0.0f, my = 0.0f;
        for (int j = ymin; j <= ymax; j++) {
            for (int i = xmin; i <= xmax; i++) {
                int hue = hsv_img.at<cv::Vec3b>(j, i)[0];
                int sat = hsv_img.at<cv::Vec3b>(j, i)[1];
                float q = hist_model.at(hue, sat);
                float p = hist_cand.at(hue, sat);

                float x_norm = (i - cx) * w_normalizer;
                float y_norm = (j - cy) * h_normalizer;
                float r2 = x_norm * x_norm + y_norm * y_norm;

                if (r2 < 1 && p != 0.0f) {
                    float w = sqrt(q / p);
                    m0 += w;
                    mx += w * i;
                    my += w * j;
                }
            }
        }
        if (m0 == 0) {
            std::cout << "m0 == 0" << std::endl;
            return 0.0f;
        }

        float cx_prev = trect.center.x;
        float cy_prev = trect.center.y;
        trect.center.x = mx / m0;
        trect.center.y = my / m0;
        displace2 =
            (trect.center.x - cx_prev) * (trect.center.x - cx_prev)
            + (trect.center.y - cy_prev) * (trect.center.y - cy_prev);

    } while (++iter < max_iter && displace2 > 1);

    float bc = hist_model.coeffBhattacharyya(hist_cand);

    double tick_end = (double)cv::getTickCount();
    printf("iter = %d\n", iter);
    printDeltaMilliSec(tick_begin, tick_end);

    return bc;
}

void visualizeLikelihood(const cv::Mat& hsv_img,
                         cv::RotatedRect trect, 
                         const HistHueSat& hist_model,
                         const HistHueSat& hist_cand,
                         cv::Mat& wmap)
{
    wmap = cv::Mat::zeros(hsv_img.rows, hsv_img.cols, CV_32F);

    int cx = (int)(trect.center.x);
    int cy = (int)(trect.center.y);
    int ww = (int)(trect.size.width / 2);
    int hh = (int)(trect.size.height / 2);
    float w_normalizer = 2 / trect.size.width;
    float h_normalizer = 2 / trect.size.height;
    int xmin = cv::max(cx - ww, 0);
    int xmax = cv::min(cx + ww, hsv_img.cols - 1);
    int ymin = cv::max(cy - hh, 0);
    int ymax = cv::min(cy + hh, hsv_img.rows - 1);

    float w_max = 0.0;
    for (int j = ymin; j <= ymax; j++) {
        for (int i = xmin; i <= xmax; i++) {

            float x_norm = (i - cx) * w_normalizer;
            float y_norm = (j - cy) * h_normalizer;
            float r2 = x_norm * x_norm + y_norm * y_norm;
            if (r2 > 1.0f) {
                continue;
            }

            int hue = hsv_img.at<cv::Vec3b>(j, i)[0];
            int sat = hsv_img.at<cv::Vec3b>(j, i)[1];
            float q = hist_model.at(hue, sat);
            float p = hist_cand.at(hue, sat);
            if (p > 0.001f) {
                float w = sqrt(q / p);
                wmap.at<float>(j, i) = w;
                if (w > w_max) {
                    w_max = w;
                }
            } else {
                wmap.at<float>(j, i) = 0.0f;
            }
        }
    }
    wmap /= w_max;
}


int main() {
    cv::namedWindow("output", cv::WINDOW_AUTOSIZE);
    cv::namedWindow("histogram", cv::WINDOW_AUTOSIZE);
    cv::namedWindow("sqrt(q/p)", cv::WINDOW_AUTOSIZE);
    cv::VideoCapture cap(0);
    cv::Mat input, input_hsv, output, wmap;
    cv::Mat disp_hist = cv::Mat(256, 256 * 3 + 2, CV_8UC3);
    disp_hist = cv::Scalar(128, 128, 128);

    HistHueSat hist_model, hist_cand;
    hist_model.visualizeLegend(disp_hist(cv::Rect(0, 0, 256, 256)));
    cv::imshow("histogram", disp_hist);

    enum TrackState tstate = TRACK_OFF;
    cv::RotatedRect trect;
    struct MouseParam mparam = { &trect, &tstate };
    cv::setMouseCallback("output", onMouse, &mparam);

    while (1) {
        cap >> input;
        cv::cvtColor(input, input_hsv, CV_BGR2HSV);
        input.copyTo(output);

        if (tstate == TRACK_INIT) {
            hist_model.calc(input_hsv, trect);
            tstate = TRACK_ON;

            hist_model.visualize(disp_hist(cv::Rect(256 + 1, 0, 256, 256)));
            cv::imshow("histogram", disp_hist);

        } else if (tstate == TRACK_SELECT
                   && trect.size.width > 0 && trect.size.height > 0) {
            cv::ellipse(output, trect, CV_RGB(255, 0, 0), 3);

        } else if (tstate == TRACK_ON) {
            float b = meanShift(input_hsv, trect, hist_model, hist_cand);
            if (b > 0.5) {
                cv::ellipse(output, trect, CV_RGB(255, 0, 0), 3);
            } else {
                cv::ellipse(output, trect, CV_RGB(0, 255, 0), 1);
            }

            hist_cand.visualize(disp_hist(cv::Rect(256 * 2 + 2, 0, 256, 256)));
            cv::imshow("histogram", disp_hist);

            visualizeLikelihood(input_hsv, trect, hist_model, hist_cand, wmap);
            cv::ellipse(wmap, trect, cv::Scalar(255), 2);
            cv::imshow("sqrt(q/p)", wmap);
        }

        cv::imshow("output", output);
        int key;
        if ((key = cv::waitKey(30)) > 0) {
            if (key == 'q' || key == 0x1b) {
                break;  
            }
        }
    }

    return 0;
}
